{% extends "base.html" %}

{% block announce %}
  <strong>that-depends</strong>  3.0 just released! If you are upgrading, check out the <a href="{{config.repo}}/migration/v3/">migration guide</a>.
{% endblock %}
